/* PTMovies	-	Display Panorama Movies
   Copyright (C) 2000 - Helmut Dersch  der@fh-furtwangen.de
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*------------------------------------------------------------*/

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.net.*;
import java.net.URLConnection;
import java.util.*;
import java.text.*;
import java.io.*;
import java.lang.reflect.*;
import java.lang.*;


public class ptzoom extends Applet implements Runnable
{
	String PTViewer = "ptviewer";
	ptviewer pv = null;
	Thread loadImages = null;
	String filename = "image";
	String maskname = null;
	int width, height;
	int idata[][] = null;
	byte hsdata[][] = null;
	private double		mt[][],md[][];			// 3 x 3 Transformation Matrix
	private int 		mi[][];					// Integer version of matrix above	
	double cpitch = 0.0, cyaw =0.0, croll = 0.0, chfov = 50.0;
	int showSHS = -1; // Show this hotspot after download is finished
	boolean progress = false; // Show progress bar during download
	Image zoom_image = null;
	String hsimage = null;
	boolean ready = false;
	int max_oversampling = 8;
	int position = 0;

	public ptzoom(){}


	public ptzoom( ptviewer p, String properties ){
		pv = p;
		setStub(new ptstub(pv, properties ));
	}	

	public ptzoom( ptviewer p, Image im, byte[] hs, String properties ){
		pv = p;
		zoom_image = im;
		if( hs != null && hs.length == im.getWidth(null)*im.getHeight(null) ){
			hsdata = new byte[im.getHeight(null)][im.getWidth(null)];
			for(int y=0; y<im.getHeight(null); y++){
				System.arraycopy(hs,y * im.getWidth(null),hsdata[y],0,im.getWidth(null));
			}
		}
		setStub(new ptstub(pv, properties ));
	}

	public ptzoom( ptviewer p, int[][] d_idata, double hfov, double pan, double tilt, double  rot){
		pv = p;
		chfov = hfov;
		cyaw = pan;
		cpitch = tilt;
		croll = rot;
	
		width = d_idata[0].length;
		height = d_idata.length;

		hsdata = new byte[height][width];
        	if(hsimage==null){
       			for(int y=0; y<height; y++)
        			for(int x=0; x<width; x++)
        				hsdata[y][x]=(byte)0xff;
		}
		
		idata = d_idata;
		setStub(new ptstub(pv, "" ));
		max_oversampling = 1000;
	}


	
   	public void init(){
   		String s;
   		
    		mt = new double[3][3];
   		md = new double[3][3];
   		mi = new int[3][3];
 
		s = getParameter( "PTViewer"); 		if( s!=null )	PTViewer = s;
 		s = getParameter( "file"); 			if( s!=null ) 	filename = s;
		s = getParameter( "mask"); 			if( s!=null ) 	maskname = s;
		s = getParameter( "hsimage"); 		if( s!=null ) 	hsimage = s;
		s = getParameter( "fov");			if( s!=null )	chfov  = Double.valueOf(s).doubleValue();
		s = getParameter( "pan");			if( s!=null )	cyaw   = Double.valueOf(s).doubleValue();
		s = getParameter( "tilt");			if( s!=null )	cpitch = Double.valueOf(s).doubleValue();
		s = getParameter( "rot");			if( s!=null )	croll  = Double.valueOf(s).doubleValue();
		s = getParameter( "showSHS"); 		if( s!=null )	showSHS  = Integer.parseInt(s);
		s = getParameter( "progress"); 		if( s!=null ) 	if( s.equalsIgnoreCase("true") ) progress = true;
		s = getParameter( "oversampling"); 	if( s!=null ) 	max_oversampling = Integer.parseInt(s);
		
	}
	
	public void start(){
		while( pv == null ){		
			try{
				pv = (ptviewer) getAppletContext().getApplet(PTViewer);
			}catch(Exception e){
   				try{ Thread.sleep( 2000 );}				
					catch (InterruptedException ie) {return;}
			}
		}

 		if( pv!=null){
 			pv.startCommunicating(this);
			if( idata == null ){
				if( zoom_image == null ){
					if( loadImages == null ){
						loadImages = new Thread( this );
						loadImages.start();
					}
				}else{
					load_zoom_image();
				}
			}else{
			       	SetDefMatrix();
 				pv.dirty = true;
        			pv.repaint();
        			ready = true; // for ptmviewer
 			}
		}
   	}

	public synchronized void set( double hfov, double pan, double tilt, double rot, int[][] data ){
		chfov = hfov;
		cyaw = pan;
		cpitch = tilt;
		croll = rot;
		
		if( data != idata ){ // New image data
			width = data[0].length;
			height = data.length;
			if(hsdata[0].length != width || hsdata.length != height){
				hsdata = new byte[height][width];
 		      			for(int y=0; y<height; y++)
        					for(int x=0; x<width; x++)
        						hsdata[y][x]=(byte)0xff;
			}
			idata = data;
		}

		SetDefMatrix();
 		pv.dirty = true;
        	pv.repaint();
	}		

	public void set( double hfov, double pan, double tilt, double rot, int[][] data, int pos ){
		position = pos;
		set( hfov, pan, tilt, rot, data );
	}
   	
   	public void stop(){
		if( pv != null ){
			if( loadImages != null ){
				pv.stopThread( loadImages );
				loadImages = null;
			}
       			if( showSHS >= 0 && showSHS < pv.numshs ){
        			pv.shs_imode[showSHS] = pv.IMODE_NORMAL;
        		}			
			pv.stopCommunicating(this);
		}
	}
	
	
	public void run(){
		load_zoom_image();
	}
	
	void load_zoom_image(){
		Image im = null;
		
		if(zoom_image == null){
			if( progress ){
				pv.ready = false;
				pv.percent[0] = 0;
				pv.repaint();
				im = pv.loadImageProgress(filename);
			}else
				im = pv.loadImage(filename);
		}else
			im = zoom_image;
			
		if(im == null) return;

        	width  = im.getWidth(null) ;
        	height = im.getHeight(null);
        
        	int[][]id;
        
        	try{
        		id = new int[height][width];
        	}catch(Exception em){
        		return;
        	}
		
		pv.ptImageTo2DArray( id, im );

        	im = null;
        
        if(maskname!=null){
        	im = pv.loadImage(maskname);
        	if(im!=null && im.getWidth(null) == width && im.getHeight(null) == height ){
        		pv.ptImageToAlpha( id, im);
        		im = null;
        	}
        }
        // Get hotspotimage
        if(hsdata == null){
        	hsdata = new byte[height][width];
        	if(hsimage==null){
        		if(pv.filename != null && pv.filename.toLowerCase().endsWith(".mov")){ // qtvr syntax
        			for(int y=0; y<height; y++)
        				for(int x=0; x<width; x++)
        					hsdata[y][x]=(byte)0;
        		}else{
       				for(int y=0; y<height; y++)
        				for(int x=0; x<width; x++)
        					hsdata[y][x]=(byte)0xff;
				}        			
        	}else{
        		im = pv.loadImage(hsimage);
        		if(im!=null && im.getWidth(null)== width && im.getHeight(null)==height){
        			int[][]data = new int[height][width];
        			pv.ptImageTo2DArray( data, im );
        			
        			for(int y=0; y<height; y++)
        			for(int x=0; x<width; x++)
        				hsdata[y][x]=(byte)(data[y][x] >> 24);
           			data = null;
        			im = null;
        			System.gc();
        		}
        	}
        }
         
        // Set up some constants
        
        SetDefMatrix();
        idata = id;
        pv.dirty = true;
        if( showSHS >=0 && showSHS<pv.numshs ){
        	pv.shs_imode[showSHS] = pv.IMODE_ALWAYS;
        }
		if( progress )
			pv.ready = true;
        pv.repaint();
        ready = true; // for ptmviewer
 	}
						       		
        		
        
	public void paint(Graphics g){
		if( pv == null || idata == null || pv.vdata == null ) return;
		
		if( pv.dirty && pv.hfov * width < max_oversampling * chfov * pv.vwidth ){
			switch( pv.quality ){
				case 0: pv.dirty = !SetView(false); break;
				case 1: if( pv.panning  || pv.lastframe>pv.frames){
							SetView(false);
						}else{
							pv.dirty = !SetView(true);
						}
						break;
				case 2: if( pv.panning ){
							SetView(false);
						}else{
							pv.dirty = !SetView(true);
						}
						break;
				case 3: pv.dirty = !SetView(true); break;
			}
		}	
	}
	

	// return true if the whole viewport is/has been rendered	
	boolean SetView( boolean bilinear ){
		int xd=0,yd=0,xs=0,ys=0,i=0,k=0;
		int v0=0,v1=0,v2=0;
		boolean result = true;
		double ta = 0.0;
		

		SetMatrix( 	pv.pitch 	* 2.0 * Math.PI / 360.0, 
				   (pv.yaw-cyaw)* 2.0 * Math.PI / 360.0, 
					1 );
		if(position == 1){
			ShiftCubeFace( (90.0) * Math.PI/180.0 );
		}else if(position == 2){
			ShiftCubeFace( -(90.0) * Math.PI/180.0 );
		}			



		double d  = ((double)pv.vwidth) / (2.0 * Math.tan((pv.hfov *Math.PI/180.0) /2.0 ));
		double ds = ((double)width) / (2.0 * Math.tan((  chfov *Math.PI/180.0) /2.0 ));

		
		for( i=0; i<3; i++){
			mt[2][i] *= d;
		}

		for(i=0; i<3; i++){
			for(k=0; k<2; k++){
				mt[i][k] *= ds;
			}
		}
		
		ta = 256.0;
		
		// Avoid integer overflow
		
		while( d * d * ta / 256.0 > 1.0e7 ) ta /= 2.0;
		
		for(i=0; i<3; i++){
			for(k=0; k<3; k++){
				mi[i][k] = (int)( ta  * mt[i][k] + 0.5);
			}
		}

		
		
		int w2 = (pv.vwidth-1) / 2;
		int h2 = pv.vheight / 2;
		int sw2 = width * 128 + 64; 
		int sh2 = height * 128 + 64;
		int h1 = height-1, w1 =width-1;
		int cy=0,dx=0,dy=0,idx=0;
		int v10, v11, v12;
		int v00, v01, v02;
		int pix;

		
		if(bilinear){
			for(yd=0, v10=(-h2)*mi[1][2]+ mi[2][2] + 128, v11=(-h2)*mi[1][0]+ mi[2][0], 
				v12=(-h2)*mi[1][1] + mi[2][1]; 
				yd<pv.vheight; 
				yd++, v10+=mi[1][2], v11+=mi[1][0], v12+=mi[1][1]){
				cy=pv.vwidth*yd;
				for(xd=0,idx=cy,v00=(-w2)*mi[0][2]+v10,v01=(-w2)*mi[0][0]+v11,
					v02=(-w2)*mi[0][1]+v12; 
				    xd<pv.vwidth; 
				    xd++,idx++){ 
					if( pv.vdata[idx] != 0 ) continue;

					v2 = (v00 + xd*mi[0][2]) / 256; //     
					if( v2 <= 0 ) {
						result = false;
						continue;
					}

					xs = (v01 + xd*mi[0][0])  / v2 + sw2 ; // center
					ys = (v02 + xd*mi[0][1])  / v2 + sh2 ;
					dx = xs & 0xff; xs = xs >> 8; 
					dy = ys & 0xff; ys = ys >> 8;

					if(xs>=0 && ys>=0 && ys<h1 && xs<w1
					   && ((pix=idata[ys][xs]) & 0xff000000) != 0){ 
						pv.vdata[idx] = pv.bil( pix, idata[ys][xs+1],
										idata[ys+1][xs], idata[ys+1][xs+1],
										dx, dy);
						pv.hs_vdata[idx] = hsdata[ys][xs];
					}else{ // Check edge pixel
						if(xs == -1) xs = 0;
						else if(xs == width) xs = w1;
						if(ys == -1) ys = 0;
						else if(ys == height) ys = h1;
						if(xs>=0 && xs<width && ys>=0 && ys<height
						 	&& ( (pix=idata[ys][xs]) & 0xff000000) != 0){
							pv.vdata[idx] = pix;
							pv.hs_vdata[idx] = hsdata[ys][xs];
						}else{
							result = false;
						}
					}
				}
			}
		}else{ // nearest neighbor
			sw2 += 128; sh2 += 128; 
			for(yd=0, v10=(-h2)*mi[1][2]+ mi[2][2] + 128, v11=(-h2)*mi[1][0]+ mi[2][0], 
				v12=(-h2)*mi[1][1] + mi[2][1]; 
				yd<pv.vheight; 
				yd++, v10+=mi[1][2], v11+=mi[1][0], v12+=mi[1][1]){
				cy=pv.vwidth*yd;
				for(xd=0,idx=cy,v00=(-w2)*mi[0][2]+v10,v01=(-w2)*mi[0][0]+v11,
					v02=(-w2)*mi[0][1]+v12; 
				    xd<pv.vwidth; 
				    xd++,idx++){ 
					if( pv.vdata[idx] != 0 ) continue;

					v2 = (v00 + xd*mi[0][2]) / 256; //     
					if( v2 <= 0 ) {
						result = false;
						continue;
					}

					xs = ((v01 + xd*mi[0][0])  / v2 + sw2) >> 8 ;
					ys = ((v02 + xd*mi[0][1])  / v2 + sh2) >> 8 ;
				
					if(xs>=0 && xs<width && ys>=0 && ys<height
						 && ( (pix=idata[ys][xs]) & 0xff000000) != 0){
						pv.vdata[idx] = pix;
					}else{ // Check edge pixel
						if(xs == -1) xs = 0;
						else if(xs == width) xs = w1;
						if(ys == -1) ys = 0;
						else if(ys == height) ys = h1;
						if(xs>=0 && xs<width && ys>=0 && ys<height
						 	&& ( (pix=idata[ys][xs]) & 0xff000000) != 0){
							pv.vdata[idx] = pix;
						}else{
							result = false;
						}
					}
				}
			}
		}
		return result;
	}
	

	void SetMatrix( double a, double b, int cl ){
		double mx[][], my[][], mz[][];
	
		mx = new double[3][3];
		my = new double[3][3];
		mz = new double[3][3];
	
		// Calculate Matrices;
		
		mx[0][0] = 1.0 ; 				mx[0][1] = 0.0 ; 				mx[0][2] = 0.0;
		mx[1][0] = 0.0 ; 				mx[1][1] = Math.cos(a) ; 		mx[1][2] = Math.sin(a);
		mx[2][0] = 0.0 ;				mx[2][1] =-mx[1][2] ;			mx[2][2] = mx[1][1];
	
		my[0][0] = Math.cos(b); 		my[0][1] = 0.0 ; 				my[0][2] =-Math.sin(b);
		my[1][0] = 0.0 ; 			my[1][1] = 1.0 ; 				my[1][2] = 0.0;
		my[2][0] = -my[0][2];			my[2][1] = 0.0 ;				my[2][2] = my[0][0];


		if( cl==1 )
			pv.matrix_matrix_mult( mx,	my,	mz);
		else
			pv.matrix_matrix_mult( my,	mx,	mz);
		pv.matrix_matrix_mult( mz, md, mt);
	}

	void SetDefMatrix(){
		double mx[][], mz[][];
	
		mx = new double[3][3];
		mz = new double[3][3];
		
		double a = (-cpitch) * 2.0 * Math.PI / 360.0; 
		double c = (- croll) * 2.0 * Math.PI / 360.0; 
	
		// Calculate Matrices;
		
		mx[0][0] = 1.0 ; mx[0][1] = 0.0 ;          mx[0][2] = 0.0;
		mx[1][0] = 0.0 ; mx[1][1] = Math.cos(a) ;  mx[1][2] = Math.sin(a);
		mx[2][0] = 0.0 ; mx[2][1] =-mx[1][2] ;	   mx[2][2] = mx[1][1];
	
		mz[0][0] = Math.cos(c) ; mz[0][1] = Math.sin(c) ; mz[0][2] = 0.0;
		mz[1][0] =-mz[0][1] ; 	 mz[1][1] = mz[0][0] ; 	  mz[1][2] = 0.0;
		mz[2][0] = 0.0 ;	 mz[2][1] = 0.0 ;	  mz[2][2] = 1.0;

		pv.matrix_matrix_mult( mz, mx, md);
	}

	public String getAppletInfo(){
		return "topFrame";
	}

	void ShiftCubeFace( double b ){
		double[][] mx = new double[3][3];
		double[][] my = new double[3][3];
		
		my[0][0] = Math.cos(b); my[0][1] = 0.0 ; my[0][2] =-Math.sin(b);
		my[1][0] = 0.0 ; my[1][1] = 1.0 ; my[1][2] = 0.0;
		my[2][0] = -my[0][2];	my[2][1] = 0.0 ;my[2][2] = my[0][0];

		pv.matrix_matrix_mult( mt, my, mx);
		for(int i=0; i<3; i++)
			for(int k=0; k<3; k++)
				mt[i][k] = mx[i][k];
	}




}
